@@ -2647,11 +2647,11 @@ tp_init_palmdetect_edge(struct tp_dispatch *tp,
 		return;
 
 	/* palm edges are 8% of the width on each side */
-	mm.x = width * 0.08;
+	mm.x = min(8, width * 0.08);
 	edges = evdev_device_mm_to_units(device, &mm);
 	tp->palm.left_edge = edges.x;
 
-	mm.x = width * 0.92;
+	mm.x = width - min(8, width * 0.08);
 	edges = evdev_device_mm_to_units(device, &mm);
 	tp->palm.right_edge = edges.x;
 
